
/* 
    Copyright 2011 Christian Pellegrin
    This software is released under the GPL v3 license
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "ftdispi.h"

static int do_debug = 0;

#define pr_debug(x...) if (do_debug) fprintf(stderr, x)

#define FTDI_DEVICE_OUT_REQTYPE ((0x02 << 5) | 0x00 | 0x00)
#define FTDI_DEVICE_IN_REQTYPE ((0x02 << 5) | 0x00 | 0x80)

#define SIO_RESET_REQUEST 0
#define SIO_SET_LATENCY_TIMER_REQUEST 0x09
#define SIO_SET_BITMODE_REQUEST       0x0B

struct ftdispi_s {
  libusb_device_handle *dev_handle;
  int mode;
  int cs_delay;
  int cs_high;
  uint8_t pin_val[2];
  uint8_t pin_dir[2];
  uint8_t *obuf;
  int obuf_n;
  uint8_t *ibuf;
  int ibuf_n;
};

static int ftdi_data_in(struct ftdispi_s *dev, uint8_t *b, int tot)
{
  int ret, actual;
  time_t start;

  if (dev->ibuf_n < (tot + 2)) {
    if (dev->ibuf)
      free(dev->ibuf);
    dev->ibuf = malloc(tot + 2);
    if (!dev->ibuf) {
      pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
      dev->ibuf_n = 0;
      return LIBUSB_ERROR_NO_MEM;
    }
    dev->ibuf_n = tot + 2;
  }

  start = time(NULL);
  do {
    ret = libusb_bulk_transfer(dev->dev_handle,
			       0x81,
			       dev->ibuf,
			       tot + 2,
			       &actual,
			       1000);
    if (ret < 0 || actual < 2) {
      pr_debug("%s %d: error %d %d/%d\n", __FUNCTION__, __LINE__, ret, actual, tot);
      return LIBUSB_ERROR_IO;
    }

    actual -= 2;
    if (actual) {
      memcpy(b, &dev->ibuf[2], actual);
      b += actual;
      tot -= actual;
    }
  } while(tot > 0 && time(NULL) <= (start + 1));

  if (tot) {
    pr_debug("%s %d: error %d %d/%d\n", __FUNCTION__, __LINE__, ret, actual, tot);
    return LIBUSB_ERROR_IO;
  }
  return 0;
}


static int ftdi_gpio_out(struct ftdispi_s *dev,
			 int heigh, uint8_t dir, uint8_t dir_mask,
			 uint8_t val, uint8_t val_mask)
{
  int ret;
  uint8_t data[3];
  int actual;

  if (heigh) {
    data[0] = 0x82;
    dev->pin_dir[1] = (dev->pin_dir[1] & ~dir_mask) | dir;
    dev->pin_val[1] = (dev->pin_val[1] & ~val_mask) | val;
    data[1] = dev->pin_val[1];
    data[2] = dev->pin_dir[1];
  }
  else {
    data[0] = 0x80;
    dev->pin_dir[0] = (dev->pin_dir[0] & ~dir_mask) | dir;
    dev->pin_val[0] = (dev->pin_val[0] & ~val_mask) | val;
    data[1] = dev->pin_val[0];
    data[2] = dev->pin_dir[0];
  }

  ret = libusb_bulk_transfer(dev->dev_handle,
			     0x02,
			     data,
			     3,
			     &actual,
			     1000);
  if (ret < 0 || actual != 3) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    if (ret >= 0)
      ret = LIBUSB_ERROR_IO;
    return ret;
  }
  return 0;
}

struct ftdispi_s *ftdispi_new(libusb_context *ctx,
			      uint16_t vendor_id, uint16_t product_id,
			      int mode,
			      int cs_delay,
			      int cs_high,
			      int *pret)
{
  int ret;
  struct ftdispi_s *dev;

  if (getenv("FTDISPI_VERBOSE"))
    do_debug = 1;

  dev = calloc(1, sizeof(*dev));
  if (!dev) {
    pr_debug("%s %d: calloc failed\n", __FUNCTION__, __LINE__);
    if (pret)
      *pret = LIBUSB_ERROR_NO_MEM;
    return NULL;
  }

  dev->mode = mode;
  dev->cs_delay = cs_delay;
  dev->cs_high = cs_high;

  dev->dev_handle = libusb_open_device_with_vid_pid(ctx, vendor_id, product_id);
  if (!dev->dev_handle) {
    pr_debug("%s %d: open device failed\n", __FUNCTION__, __LINE__);
    if (pret)
      *pret = LIBUSB_ERROR_NO_DEVICE;
    free(dev);
    return NULL;
  }

  /* reset the FTDI */
  ret = libusb_control_transfer(dev->dev_handle,
				FTDI_DEVICE_OUT_REQTYPE,
				SIO_RESET_REQUEST,
				0,
				0,
				NULL,
				0,
				1000);
  if (ret < 0) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    if (pret)
      *pret = ret;
    free(dev);
    return NULL;
  }

  usleep(100*1000);		/* wait for reset to finish */

  /* set to the minimum latency */
  ret = libusb_control_transfer(dev->dev_handle,
				FTDI_DEVICE_OUT_REQTYPE,
				SIO_SET_LATENCY_TIMER_REQUEST,
				2,
				0,
				NULL,
				0,
				1000);
  if (ret < 0) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    if (pret)
      *pret = ret;
    free(dev);
    return NULL;
  }

  /* set to mpsse mode */
  ret = libusb_control_transfer(dev->dev_handle,
				FTDI_DEVICE_OUT_REQTYPE,
				SIO_SET_BITMODE_REQUEST,
				0x020b,
				0,
				NULL,
				0,
				1000);
  if (ret < 0) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    if (pret)
      *pret = ret;
    free(dev);
    return NULL;
  }

  ret = ftdi_gpio_out(dev,
		      0, 0xb, 0xff,
		      (mode & 2 ? 1 : 0) | (cs_high ? 0 : 8),
		      0xff);
  if (ret < 0) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    if (pret)
      *pret = ret;
    free(dev);
    return NULL;
  }

  /* set gpio for high port */
  ret = ftdi_gpio_out(dev, 1, 0x0, 0xff, 0x0, 0xff);
  if (ret < 0) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    if (pret)
      *pret = ret;
    free(dev);
    return NULL;
  }

  return dev;
}

int ftdispi_spi(struct ftdispi_s *dev,
		uint8_t  *data_out,
		int n_out,
		uint8_t  *data_in,
		int n_in)
{
  int ret = 0;
#define CHUNK 64
  int actual, chunk, done;

  if (dev->obuf_n < (n_out + 3)) {
    if (dev->obuf)
      free(dev->obuf);
    dev->obuf = malloc(n_out + 3);
    if (!dev->obuf) {
      pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
      dev->obuf_n = 0;
      return LIBUSB_ERROR_NO_MEM;
    }
    dev->obuf_n = n_out + 3;
  }

  ret = ftdi_gpio_out(dev, 0, 0, 0, dev->cs_high ? 0x08 : 0, 0x08);
  if (ret < 0) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    return ret;
  }
  usleep(dev->cs_delay);

  done = 0;
  while (done < n_out) {
    chunk = n_out - done;
    if (chunk > CHUNK)
      chunk = CHUNK;

    dev->obuf[0] = 0x10 | (n_in > 0 ? 0x20 : 0);
    switch (dev->mode) {
    case 1: dev->obuf[0] |= 0x04; break;
    case 2: dev->obuf[0] |= 0x04; break;
    case 3: dev->obuf[0] |= 0x01; break;
    default: dev->obuf[0] |= 0x01; break;
    }
    dev->obuf[1] = (chunk - 1) & 0xff;
    dev->obuf[2] = ((chunk - 1) >> 8)  & 0xff;
    memcpy(&dev->obuf[3], &data_out[done], chunk);

    ret = libusb_bulk_transfer(dev->dev_handle,
			       0x02,
			       dev->obuf,
			       3 + chunk,
			       &actual,
			       1000);
    if (ret < 0 || actual != (3 + chunk)) {
      pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
      if (ret >= 0)
	ret = LIBUSB_ERROR_IO;
      goto io_finished;
    }

    if (n_in > 0) {
      ret = ftdi_data_in(dev, &data_in[done], chunk);
      if (ret) {
	pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
	goto io_finished;
      }
    }

    done += chunk;
  }

  usleep(dev->cs_delay);
 io_finished:
  ftdi_gpio_out(dev, 0, 0, 0, dev->cs_high ? 0 : 0x08, 0x08);
  return ret;
}

int ftdispi_gpio_conf(struct ftdispi_s *dev,
		      int gpio, int dir, int val)
{
  int mask = 1 << (gpio & 0xf);

  return ftdi_gpio_out(dev, gpio & 0x10, dir ? mask : 0, mask, val ? mask : 0, mask);
}

int ftdispi_gpio_set(struct ftdispi_s *dev,
		     int gpio, int val)
{
  int mask = 1 << (gpio & 0xf);

  return ftdi_gpio_out(dev, gpio & 0x10, 0, 0, val ? mask : 0, mask);
}

int ftdispi_gpio_get(struct ftdispi_s *dev,
		     int gpio)
{
  int ret;
  uint8_t data[1];
  int actual;

  data[0] = (gpio & 0x10) ? 0x81 : 0x83;
  ret = libusb_bulk_transfer(dev->dev_handle,
			     0x02,
			     data,
			     1,
			     &actual,
			     1000);
  if (ret < 0 || actual != 1) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    if (ret >= 0)
      ret = LIBUSB_ERROR_IO;
    return ret;
  }

  ret = ftdi_data_in(dev, data, 1);
  if (ret) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    return ret;
  }

  return (data[0] & (gpio & 0xf)) > 0;
}

int ftdispi_loop(struct ftdispi_s *dev, int on)
{
  int ret;
  uint8_t data[1];
  int actual;

  data[0] = on ? 0x84 : 0x85;
  ret = libusb_bulk_transfer(dev->dev_handle,
			     0x02,
			     data,
			     1,
			     &actual,
			     1000);
  if (ret < 0 || actual != 1) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    if (ret >= 0)
      ret = LIBUSB_ERROR_IO;
    return ret;
  }
  return 0;
}

int ftdispi_freq(struct ftdispi_s *dev, unsigned long hz)
{
  unsigned long actual_hz;
  unsigned long div;
  int ret;
  uint8_t data[3];
  int actual;

  div = 6 * 1000 * 1000 / hz;
  if (div == 0)
    div = 1;
  if (div > 65536)
    div = 65537;
  div -= 1;

  actual_hz = 6 * 1000 * 1000 / (div + 1);
  if (hz != actual_hz)
    pr_debug("speed %lu requested %lu Hz\n", actual_hz, hz);

  data[0] = 0x86;
  data[1] = div & 0xff;
  data[2] = (div >> 8) & 0xff;
  ret = libusb_bulk_transfer(dev->dev_handle,
			     0x02,
			     data,
			     3,
			     &actual,
			     1000);
  if (ret < 0 || actual != 3) {
    pr_debug("%s %d: error %d\n", __FUNCTION__, __LINE__, ret);
    if (ret >= 0)
      ret = LIBUSB_ERROR_IO;
    return ret;
  }
  return 0;
}

void ftdispi_delete(struct ftdispi_s *dev)
{
  libusb_close(dev->dev_handle);
  if (dev->obuf)
    free(dev->obuf);
  if (dev->ibuf)
    free(dev->ibuf);
  free(dev);
}

